# Copyright 2021 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from pathlib import Path
import tempfile
from typing import List, Sequence, Tuple, Union

import nmslib
import tensorflow as tf

from .search import Search
from tensorflow_similarity.distances import Distance, distance_canonicalizer
from tensorflow_similarity.types import FloatTensor


class NMSLibSearch(Search):
    """
    Efficiently find nearest embeddings by indexing known embeddings and make
    them searchable using the  [Approximate Nearest Neigboors Search](https://en.wikipedia.org/wiki/Nearest_neighbor_search)
    search library [NMSLIB](https://github.com/nmslib/nmslib).
    """

    def __init__(
        self,
        distance: Union[Distance, str],
        dims: int,
        algorithm: str = "nmslib_hnsw",
        **kwargs
    ):

        distance_obj: Distance = distance_canonicalizer(distance)

        # convert to nmslib word
        if distance_obj.name == "cosine":
            space = "cosinesimil"
        elif distance_obj.name == "euclidean":
            space = "l2"
        elif distance_obj.name == "manhattan":
            space = "l1"
        else:
            raise ValueError("Unsupported metric space")

        if algorithm == "nmslib_hnsw":
            method = "hnsw"
        else:
            raise ValueError("Unsupported algorithm")

        self._search_index = nmslib.init(method=method, space=space)
        self._search_index.createIndex()

    def add(
        self,
        embedding: FloatTensor,
        idx: int,
        verbose: int = 1,
        build: bool = True,
        **kwargs
    ):
        """Add an embedding to the index

        Args:
            embedding: The embedding to index as computed by
            the similarity model.

            idx: Embedding id as in the index table.
            Returned with the embedding to allow to lookup
            the data associated with a given embedding.

            verbose: Be verbose. Defaults to 1.

            build: Rebuild the index after the addition.
            Required to make the embedding searchable.
            Set to false to save time between successive addition.
            Defaults to True.

        """
        self._search_index.addDataPoint(idx, embedding)
        if build:
            self._build(verbose=verbose)

    def batch_add(
        self,
        embeddings: FloatTensor,
        idxs: Sequence[int],
        verbose: int = 1,
        build: bool = True,
        **kwargs
    ):
        """Add a batch of embeddings to the search index.

        Args:
            embeddings: List of embeddings to add to the index.

            idxs (int): Embedding ids as in the index table. Returned with
            the embeddings to allow to lookup the data associated
            with the returned embeddings.

            verbose: Be verbose. Defaults to 1.

            build: Rebuild the index after the addition. Required to
            make the embeddings searchable. Set to false to save
            time between successive addition. Defaults to True.
        """
        # !addDataPoint and addDataPointBAtch have inverted parameters
        if verbose:
            print("|-Adding embeddings to index.")
        self._search_index.addDataPointBatch(embeddings, idxs)

        if build:
            if verbose:
                print("|-Building index.")
            self._build(verbose=verbose)

    def lookup(
        self, embedding: FloatTensor, k: int = 5
    ) -> Tuple[List[int], List[float]]:
        """Find embedding K nearest neighboors embeddings.

        Args:
            embedding: Query embedding as predicted by the model.
            k: Number of nearest neighboors embedding to lookup. Defaults to 5.
        """
        idxs: List[int] = []
        distances: List[float] = []
        idxs, distances = self._search_index.knnQuery(embedding, k=k)
        return idxs, distances

    def batch_lookup(
        self, embeddings: FloatTensor, k: int = 5
    ) -> Tuple[List[List[int]], List[List[float]]]:
        """Find embeddings K nearest neighboors embeddings.

        Args:
            embedding: Batch of query embeddings as predicted by the model.
            k: Number of nearest neighboors embedding to lookup. Defaults to 5.
        """
        batch_idxs = []
        batch_distances = []

        nn = self._search_index.knnQueryBatch(embeddings, k=k)
        for n in nn:
            batch_idxs.append(n[0])
            batch_distances.append(n[1])
        return batch_idxs, batch_distances

    def save(self, path: str):
        """Serializes the index data on disk

        Args:
            path: where to store the data
        """
        with tempfile.TemporaryDirectory() as tmpdirname:
            fname = self.__make_fname(path)
            tmpidx = os.path.join(tmpdirname, os.path.basename(path))
            self._search_index.saveIndex(tmpidx, save_data=True)

            # read search_index.bin tmp file and write to fname
            with tf.io.gfile.GFile(fname, "w+b") as gfp:
                with open(tmpidx, "rb") as fp:
                    gfp.write(fp.read())

            # read search_index.bin.dat tmp file and write to fname + .dat
            with tf.io.gfile.GFile(fname + ".dat", "w+b") as gfp:
                tmpdat = os.path.join(
                    tmpdirname, os.path.basename(path) + ".dat"
                )
                with open(tmpdat, "rb") as fp:
                    gfp.write(fp.read())

    def load(self, path: str):
        """load index on disk

        Args:
            path: where to store the data
        """
        with tempfile.TemporaryDirectory() as tmpdirname:
            fname = self.__make_fname(path)
            tmpidx = os.path.join(tmpdirname, os.path.basename(path))

            # read fname and write to tmpidx
            with tf.io.gfile.GFile(fname, "rb") as gfp:
                with open(tmpidx, "w+b") as fp:
                    fp.write(gfp.read())

            # read fname + .dat and write to tmpdat
            with tf.io.gfile.GFile(fname + ".dat", "rb") as gfp:
                tmpdat = os.path.join(
                    tmpdirname, os.path.basename(path) + ".dat"
                )
                with open(tmpdat, "w+b") as fp:
                    fp.write(gfp.read())

            self._search_index.loadIndex(tmpidx, load_data=True)

    def _build(self, verbose=0):
        """Build the index this is need to take into account the new points"""
        show = True if verbose else False
        self._search_index.createIndex(print_progress=show)

    def __make_fname(self, path):
        return str(Path(path) / "search_index.bin")



import os

import tensorflow as tf

from tensorflow_similarity.samplers import TFRecordDatasetSampler


def _int64_feature(value):
    """Returns an int64_list from a bool / enum / int / uint."""
    return tf.train.Feature(int64_list=tf.train.Int64List(value=[value]))


def to_tfrecord(sid, value):
    feature = {
        "sid": _int64_feature(sid),
        "value": _int64_feature(value),
    }

    example = tf.train.Example(features=tf.train.Features(feature=feature))
    return example.SerializeToString()


def deserialization_fn(serialized_example):
    fd = {
        "sid": tf.io.FixedLenFeature([], dtype=tf.int64),
        "value": tf.io.FixedLenFeature([], dtype=tf.int64),
    }
    sample = tf.io.parse_single_example(serialized_example, fd)

    return (sample["sid"], sample["value"])


class TFRecordSamplerTest(tf.test.TestCase):
    def setUp(self):
        super().setUp()

        for sid in range(100):
            shard_path = os.path.join(self.get_temp_dir(), f"tfr_{sid}.tfrec")

            with tf.io.TFRecordWriter(str(shard_path)) as w:
                for value in range(1000):
                    example = to_tfrecord(sid, sid * 1000 + value)
                    w.write(example)

    def test_basic(self):
        sampler = TFRecordDatasetSampler(
            self.get_temp_dir(),
            deserialization_fn=deserialization_fn,
            batch_size=10,
            example_per_class=2,
        )

        si = iter(sampler)
        [next(si) for _ in range(10_000)]
        examples = next(si)

        # We should get 3 pairs of shard IDs
        sids = examples[0]
        values = examples[1]
        first_sid = sids[::2]
        second_sid = sids[1::2]

        self.assertLen(sids, 10)
        self.assertAllEqual(first_sid, second_sid)

        for sid, val in zip(sids, values):
            diff = val - sid * 1_000
            self.assertGreaterEqual(diff, 0)
            self.assertLess(diff, 1000)



# Copyright 2021 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"EfficientNet backbone for similarity learning"
import re
from typing import Tuple
from tensorflow.keras import layers
from tensorflow.keras.applications import efficientnet
from tensorflow_similarity.layers import MetricEmbedding
from tensorflow_similarity.layers import GeneralizedMeanPooling2D
from tensorflow_similarity.models import SimilarityModel

EFF_INPUT_SIZE = {
    "B0": 224,
    "B1": 240,
    "B2": 260,
    "B3": 300,
    "B4": 380,
    "B5": 456,
    "B6": 528,
    "B7": 600,
}

EFF_ARCHITECTURE = {
    "B0": efficientnet.EfficientNetB0,
    "B1": efficientnet.EfficientNetB1,
    "B2": efficientnet.EfficientNetB2,
    "B3": efficientnet.EfficientNetB3,
    "B4": efficientnet.EfficientNetB4,
    "B5": efficientnet.EfficientNetB5,
    "B6": efficientnet.EfficientNetB6,
    "B7": efficientnet.EfficientNetB7,
}


def EfficientNetSim(
    input_shape: Tuple[int, int, int],
    embedding_size: int = 128,
    variant: str = "B0",
    weights: str = "imagenet",
    trainable: str = "frozen",
    l2_norm: bool = True,
    include_top: bool = True,
    pooling: str = "gem",
    gem_p=3.0,
) -> SimilarityModel:
    """Build an EfficientNet Model backbone for similarity learning

    [EfficientNet: Rethinking Model Scaling for Convolutional Neural Networks](https://arxiv.org/abs/1905.11946)

    Args:
        input_shape: Size of the input image. Must match size of EfficientNet version you use.
        See below for version input size.

        embedding_size: Size of the output embedding. Usually between 64
        and 512. Defaults to 128.

        variant: Which Variant of the EfficientNet to use. Defaults to "B0".

        weights: Use pre-trained weights - the only available currently being
        imagenet. Defaults to "imagenet".

        trainable: Make the EfficientNet backbone fully trainable or partially
        trainable.
        - "full" to make the entire backbone trainable,
        - "partial" to only make the last 3 block trainable
        - "frozen" to make it not trainable.

        l2_norm: If True and include_top is also True, then
        tfsim.layers.MetricEmbedding is used as the last layer, otherwise
        keras.layers.Dense is used. This should be true when using cosine
        distance. Defaults to True.

        include_top: Whether to include the fully-connected layer at the top
        of the network. Defaults to True.

        pooling: Optional pooling mode for feature extraction when
        include_top is False. Defaults to gem.
        - None means that the output of the model will be the 4D tensor
          output of the last convolutional layer.
        - avg means that global average pooling will be applied to the
          output of the last convolutional layer, and thus the output of the
          model will be a 2D tensor.
        - max means that global max pooling will be applied.
        - gem means that global GeneralizedMeanPooling2D will be applied.
          The gem_p param sets the contrast amount on the pooling.

        gem_p: Sets the power in the GeneralizedMeanPooling2D layer. A value
        of 1.0 is equivalent to GlobalMeanPooling2D, while larger values
        will increase the contrast between activations within each feature
        map, and a value of math.inf will be equivalent to MaxPool2d.

    Note:
        EfficientNet expects images at the following size:
         - "B0": 224,
         - "B1": 240,
         - "B2": 260,
         - "B3": 300,
         - "B4": 380,
         - "B5": 456,
         - "B6": 528,
         - "B7": 600,

    """

    # input
    inputs = layers.Input(shape=input_shape)
    x = inputs

    if variant not in EFF_INPUT_SIZE:
        raise ValueError("Unknown efficientnet variant. Valid B0...B7")

    x = build_effnet(x, variant, weights, trainable)

    if pooling == "gem":
        x = GeneralizedMeanPooling2D(p=gem_p, name="gem_pool")(x)
    elif pooling == "avg":
        x = layers.GlobalAveragePooling2D(name="avg_pool")(x)
    elif pooling == "max":
        x = layers.GlobalMaxPooling2D(name="max_pool")(x)

    if include_top and pooling is not None:
        if l2_norm:
            outputs = MetricEmbedding(embedding_size)(x)
        else:
            outputs = layers.Dense(embedding_size)(x)
    else:
        outputs = x

    return SimilarityModel(inputs, outputs)


def build_effnet(
    x: layers.Layer, variant: str, weights: str, trainable: str
) -> layers.Layer:
    """Build the requested efficient net.

    Args:
        x: The input layer to the efficientnet.

        variant: Which Variant of the EfficientNet to use.

        weights: Use pre-trained weights - the only available currently being
        imagenet.

        trainable: Make the EfficientNet backbone fully trainable or partially
        trainable.
        - "full" to make the entire backbone trainable,
        - "partial" to only make the last 3 block trainable
        - "frozen" to make it not trainable.

    Returns:
        The output layer of the efficientnet model
    """

    # init
    effnet_fn = EFF_ARCHITECTURE[variant.upper()]
    effnet = effnet_fn(weights=weights, include_top=False)

    if trainable == "full":
        effnet.trainable = True
    elif trainable == "partial":
        # let's mark the top part of the network as trainable
        effnet.trainable = True
        for layer in effnet.layers:
            # Freeze all the layers before the the last 3 blocks
            if not re.search("^block[5,6,7]|^top", layer.name):
                layer.trainable = False
    elif trainable == "frozen":
        effnet.trainable = False
    else:
        raise ValueError(
            f"{trainable} is not a supported option for 'trainable'."
        )

    # Don't train the BN layers if we are loading pre-trained weights.
    if weights:
        for layer in effnet.layers:
            if isinstance(layer, layers.BatchNormalization):
                layer.trainable = False

    # wire
    x = effnet(x)

    return x



# Copyright 2021 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"Various utilities functions for improved quality of life."
import math
from typing import Dict, List, Optional, Sequence, Union

import numpy as np
import tensorflow as tf

from tensorflow_similarity.types import BoolTensor
from tensorflow_similarity.types import FloatTensor
from tensorflow_similarity.types import IntTensor
from tensorflow_similarity.types import Lookup


def is_tensor_or_variable(x):
    "check if a variable is tf.Tensor or tf.Variable"
    return tf.is_tensor(x) or isinstance(x, tf.Variable)


def tf_cap_memory():
    "Avoid TF to hog memory before needing it"
    gpus = tf.config.experimental.list_physical_devices("GPU")

    if gpus:
        for gpu in gpus:
            try:
                tf.config.experimental.set_memory_growth(gpu, True)
            except RuntimeError as e:
                # Memory growth must be set before GPUs have been initialized
                print(e)


def unpack_lookup_labels(lookups: Sequence[Sequence[Lookup]], dtype: Union[str, tf.DType]) -> IntTensor:
    # Using list comprehension as it is faster
    all_values = [[n.label for n in lu] for lu in lookups]
    # Lookup sets are not guaranteed to all be the same size. Therefore we load
    # the list of lists as a ragged tensor and convert to an int tensor with a
    # default class label value set to the max value for int.
    base_type = tf.dtypes.as_dtype(dtype).base_dtype
    ragged_labels = tf.ragged.constant(all_values, dtype=base_type)

    if not _same_length_rows(ragged_labels):
        print(f"WARNING: {_count_of_small_lookup_sets(ragged_labels)} lookup "
              "sets are shorter than the max lookup set length. Imputing "
              "0x7FFFFFFF for the missing label lookups.")

    result: IntTensor = ragged_labels.to_tensor(default_value=0x7FFFFFFF)

    return result


def unpack_lookup_distances(
        lookups: Sequence[Sequence[Lookup]],
        dtype: Union[str, tf.DType],
        distance_rounding: Optional[int] = None) -> FloatTensor:
    # using list comprehension as it is faster
    all_values = [[n.distance for n in lu] for lu in lookups]
    # Lookup sets are not guaranteed to all be the same size. Therefore we load
    # the list of lists as a ragged tensor and convert to an flat32 tensor with a
    # default dist value set to math.inf.
    base_type = tf.dtypes.as_dtype(dtype).base_dtype
    ragged_dists = tf.ragged.constant(all_values, dtype=base_type)

    if distance_rounding is not None:
        multiplier = tf.constant([10.0**distance_rounding])
        ragged_dists = tf.round(ragged_dists * multiplier) / multiplier

    if not _same_length_rows(ragged_dists):
        print(f"WARNING: {_count_of_small_lookup_sets(ragged_dists)} lookup "
              "sets are shorter than the max lookup set length. Imputing "
              "math.inf for the missing distance lookups.")

    dists: FloatTensor = ragged_dists.to_tensor(default_value=math.inf)

    return dists


def unpack_results(results: Dict[str, np.ndarray], epoch: int, logs: dict,
                   tb_writer: tf.summary.SummaryWriter,
                   name_suffix: Optional[str] = "") -> List[str]:
    """Updates logs, writes summary, and returns list of strings of
    evaluation metric"""
    mstr = []
    for metric_name, vals in results.items():
        float_val = vals[0]
        full_metric_name = f"{metric_name}{name_suffix}"
        logs[full_metric_name] = float_val
        mstr.append(f"{full_metric_name}: {float_val:.4f}")
        if tb_writer:
            with tb_writer.as_default():
                tf.summary.scalar(full_metric_name, float_val, step=epoch)
    return mstr


def _same_length_rows(x: tf.RaggedTensor) -> BoolTensor:
    """Check if the rows are all the same length.

    Args:
        x: A RaggedTensor

    Returns:
       True if all rows are the same length.
    """
    dims = tf.expand_dims(x.row_lengths(), axis=-1)
    pairwise_equality = tf.equal(dims, tf.transpose(dims))
    is_same_length: BoolTensor = tf.math.reduce_all(pairwise_equality)
    return is_same_length


def _count_of_small_lookup_sets(x: tf.RaggedTensor) -> IntTensor:
    """The count of lookup sets smaller than x.bounding_shape()[1]

    Args:
        x: A RaggedTensor

    Returns:
        The count of smaller lookup sets.
    """
    rl = x.row_lengths()
    max_rl = x.bounding_shape()[1]
    short_lookup_sets = tf.cast(rl != max_rl, dtype="int32")
    small_lookup_count: IntTensor = tf.math.reduce_sum(short_lookup_sets)
    return small_lookup_count



# Copyright 2020 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Barlow Loss.
    Barlow Twins: Self-Supervised Learning via Redundancy Reduction
    https://arxiv.org/abs/2103.03230
"""
from typing import Any, Callable, Dict, Optional

import tensorflow as tf

from tensorflow_similarity.types import FloatTensor


@tf.keras.utils.register_keras_serializable(package="Similarity")
class Barlow(tf.keras.losses.Loss):
    """Computes the Barlow Loss between two batches of embeddings.

    Reference

    Zbontar, Jure, et al.
    "Barlow Twins: Self-Supervised Learning via Redundancy Reduction."
    https://arxiv.org/abs/2103.03230

    Standalone usage:

    >>> loss = tensorflow_similarity.losses.Barlow()
    >>> za = tf.random.uniform(shape=[4, 16])
    >>> zb = tf.random.uniform(shape=[4, 16])
    >>> loss(za, zb)
    <tf.Tensor: shape=(), dtype=float32, numpy=22.144062>

    Usage with the `compile()` API:
    ```python
    model.compile(optimizer='sgd', loss=tensorflow_similarity.losses.Barlow())
    ```
    """

    def __init__(
        self,
        lambda_: float = 5e-3,
        margin: float = 1e-12,
        reduction: Callable = tf.keras.losses.Reduction.AUTO,
        name: Optional[str] = None,
        **kwargs
    ):
        super().__init__(reduction=reduction, name=name, **kwargs)
        self.lambda_ = lambda_
        self.margin = margin

    def call(self, za: FloatTensor, zb: FloatTensor) -> FloatTensor:
        """Compute the lost.

        Args:
            za: Embedding A
            zb: Embedding B

        Returns:
            loss
        """
        # compute the diagonal
        batch_size = tf.shape(za)[0]

        za = self.standardize_columns(za)
        zb = self.standardize_columns(zb)

        # compute pairwise
        c = tf.matmul(za, zb, transpose_a=True)
        c = c / tf.cast(batch_size, dtype="float32")

        on_diag = 1.0 - tf.linalg.diag_part(c)
        on_diag = tf.math.pow(on_diag, 2)
        on_diag = tf.math.reduce_sum(on_diag)

        off_diag = self.off_diagonal(c)
        off_diag = tf.math.pow(off_diag, 2)
        off_diag = tf.math.reduce_sum(off_diag)

        # 1D Tensor
        loss: FloatTensor = off_diag * self.lambda_ + on_diag + self.margin

        return loss

    def get_config(self) -> Dict[str, Any]:
        config = {
            "lambda_": self.lambda_,
            "margin": self.margin,
        }
        base_config = super().get_config()
        return {**base_config, **config}

    def off_diagonal(self, x: FloatTensor) -> FloatTensor:
        n = tf.shape(x)[0]
        flattened = tf.reshape(x, [-1])[:-1]
        off_diagonals = tf.reshape(flattened, (n - 1, n + 1))[:, 1:]
        off_diag: FloatTensor = tf.reshape(off_diagonals, [-1])
        return off_diag

    def standardize_columns(self, x: FloatTensor) -> FloatTensor:
        col_mean = tf.math.reduce_mean(x, axis=0)
        col_std = tf.math.reduce_std(x, axis=0)

        norm_col: FloatTensor = tf.math.divide_no_nan((x - col_mean), col_std)
        return norm_col



from typing import List

import tensorflow as tf

from tensorflow_similarity.types import FloatTensor, Tensor
from tensorflow_similarity.augmenters.augmentation_utils.random_apply import (
    random_apply,
)


def random_blur(
    image: Tensor,
    height: int,
    width: int,
    p: float = 1.0,
    min_sigma: float = 0.1,
    max_sigma: float = 2.0,
    kernel_size_divider: float = 10,
) -> Tensor:
    """Randomly blur an image.

    Args:
      image: `Tensor` representing an image of arbitrary size.
      height: Height of output image.
      width: Width of output image.
      p: probability of applying this transformation.

    Returns:
      A preprocessed image `Tensor`.
    """
    del width

    def _transform(image: Tensor) -> Tensor:
        sigma = tf.random.uniform([], min_sigma, max_sigma, dtype=tf.float32)
        return gaussian_blur(
            image,
            kernel_size=height // kernel_size_divider,
            sigma=sigma,
            padding="SAME",
        )

    return random_apply(_transform, p=p, x=image)


def batch_random_blur(
    images_list: List[Tensor],
    height: int,
    width: int,
    blur_probability: float = 0.5,
) -> List[Tensor]:
    """Apply efficient batch data transformations.

    Args:
      images_list: a list of image tensors.
      height: the height of image.
      width: the width of image.
      blur_probability: the probaility to apply the blur operator.

    Returns:
      Preprocessed feature list.
    """

    def generate_selector(p: float, bsz: int) -> FloatTensor:
        shape = [bsz, 1]
        cond = tf.less(tf.random.uniform(shape, 0, 1, dtype=tf.float32), p)
        selector: FloatTensor = tf.cast(cond, tf.float32)
        return selector

    new_images_list = []
    for images in images_list:
        images_new = random_blur(images, height, width, p=1.0)
        selector = generate_selector(blur_probability, tf.shape(images)[0])
        images = images_new * selector + images * (1 - selector)
        images = tf.clip_by_value(images, 0.0, 1.0)
        new_images_list.append(images)

    return new_images_list


def gaussian_blur(
    image: Tensor,
    kernel_size: Tensor,
    sigma: float,
    padding: str = "SAME",
) -> Tensor:
    """Blurs the given image with separable convolution.


    Args:
      image: Tensor of shape [height, width, channels] and dtype float to blur.
      kernel_size: Integer Tensor for the size of the blur kernel. This is
      should be an odd number. If it is an even number, the actual kernel
      size will besize + 1.
      sigma: Sigma value for gaussian operator.
      padding: Padding to use for the convolution. Typically 'SAME' or 'VALID'.

    Returns:
      A Tensor representing the blurred image.
    """
    radius = tf.cast(kernel_size / 2, dtype=tf.int32)
    kernel_size = radius * 2 + 1
    x = tf.cast(tf.range(-radius, radius + 1), dtype=tf.float32)
    blur_filter = tf.exp(
        -tf.pow(x, 2.0) / (2.0 * tf.pow(tf.cast(sigma, dtype=tf.float32), 2.0))
    )
    blur_filter /= tf.reduce_sum(blur_filter)
    # One vertical and one horizontal filter.
    blur_v = tf.reshape(blur_filter, [kernel_size, 1, 1, 1])
    blur_h = tf.reshape(blur_filter, [1, kernel_size, 1, 1])
    num_channels = tf.shape(image)[-1]
    blur_h = tf.tile(blur_h, [1, 1, num_channels, 1])
    blur_v = tf.tile(blur_v, [1, 1, num_channels, 1])
    expand_batch_dim = image.shape.ndims == 3
    if expand_batch_dim:
        # Tensorflow requires batched input to convolutions,
        # which we can fake with an extra dimension.
        image = tf.expand_dims(image, axis=0)
    blurred = tf.nn.depthwise_conv2d(
        image, blur_h, strides=[1, 1, 1, 1], padding=padding
    )
    blurred = tf.nn.depthwise_conv2d(
        blurred, blur_v, strides=[1, 1, 1, 1], padding=padding
    )
    if expand_batch_dim:
        blurred = tf.squeeze(blurred, axis=0)
    return blurred



from typing import Callable

import tensorflow as tf

from tensorflow_similarity.types import Tensor


def random_apply(
    func: Callable[[Tensor], Tensor], p: float, x: Tensor
) -> Tensor:
    """Randomly apply function func to x with probability p."""
    return tf.cond(
        tf.less(
            tf.random.uniform([], minval=0, maxval=1, dtype=tf.float32),
            tf.cast(p, tf.float32),
        ),
        lambda: func(x),
        lambda: x,
    )



# Copyright 2021 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"ResNet50 backbone for similarity learning"
from typing import Tuple
import tensorflow as tf
from tensorflow.keras import layers
from tensorflow_similarity.layers import MetricEmbedding
from tensorflow_similarity.layers import GeneralizedMeanPooling2D
from tensorflow_similarity.models import SimilarityModel


# Create an image augmentation pipeline.
def ResNet18Sim(
    input_shape: Tuple[int, int, int],
    embedding_size: int = 128,
    l2_norm: bool = True,
    include_top: bool = True,
    pooling: str = "gem",
    gem_p=3.0,
) -> SimilarityModel:
    """Build an ResNet18 Model backbone for similarity learning

    Architecture from [Deep Residual Learning for Image Recognition](https://arxiv.org/abs/1512.03385)

    Args:
        input_shape: Expected to be betweeen 32 and 224 and in the (H, W, C)
        data_format.

        embedding_size: Size of the output embedding. Usually between 64
        and 512. Defaults to 128.

        l2_norm: If True and include_top is also True, then
        tfsim.layers.MetricEmbedding is used as the last layer, otherwise
        keras.layers.Dense is used. This should be true when using cosine
        distance. Defaults to True.

        include_top: Whether to include a fully-connected layer of
        embedding_size at the top of the network. Defaults to True.

        pooling: Optional pooling mode for feature extraction when
        include_top is False. Defaults to gem.
        - None means that the output of the model will be the 4D tensor
          output of the last convolutional layer.
        - avg means that global average pooling will be applied to the
          output of the last convolutional layer, and thus the output of the
          model will be a 2D tensor.
        - max means that global max pooling will be applied.
        - gem means that global GeneralizedMeanPooling2D will be applied.
          The gem_p param sets the contrast amount on the pooling.

        gem_p: Sets the power in the GeneralizedMeanPooling2D layer. A value
        of 1.0 is equivelent to GlobalMeanPooling2D, while larger values
        will increase the contrast between activations within each feature
        map, and a value of math.inf will be equivelent to MaxPool2d.
    """

    # input
    inputs = layers.Input(shape=input_shape)
    x = inputs

    x = build_resnet(x, input_shape)

    if pooling == "gem":
        x = GeneralizedMeanPooling2D(p=gem_p, name="gem_pool")(x)
    elif pooling == "avg":
        x = layers.GlobalAveragePooling2D(name="avg_pool")(x)
    elif pooling == "max":
        x = layers.GlobalMaxPooling2D(name="max_pool")(x)

    if include_top and pooling is not None:
        if l2_norm:
            outputs = MetricEmbedding(embedding_size)(x)
        else:
            outputs = layers.Dense(embedding_size)(x)
    else:
        outputs = x

    return SimilarityModel(inputs, outputs, name="resnet18sim")


def build_resnet(
    x: layers.Layer, input_shape: Tuple[int, int, int]
) -> layers.Layer:
    """Build the requested ResNet.

    Args:
        x: The input layer to the ResNet.

        input_shape: Expected to be betweeen 32 and 224 and in the (H, W, C)
        data_format.

    Returns:
        The ouptut layer of the ResNet model
    """
    inputs = layers.Input(shape=input_shape)

    layer = tf.keras.layers.ZeroPadding2D(
        padding=((1, 1), (1, 1)), name="conv1_pad"
    )(inputs)
    layer = tf.keras.layers.Conv2D(
        64,
        kernel_size=3,
        strides=1,
        use_bias=False,
        kernel_initializer=tf.keras.initializers.LecunUniform(),
        name="conv1_conv",
    )(layer)
    layer = tf.keras.layers.BatchNormalization(
        epsilon=1.001e-5, name="conv1_bn"
    )(layer)
    layer = tf.keras.layers.Activation("relu", name="conv1_relu")(layer)

    outputs = stack_fn(layer)

    # wire
    x = tf.keras.Model(inputs, outputs, name="resnet18")(x)

    return x


def block0(
    x,
    filters,
    kernel_size=3,
    stride=1,
    conv_shortcut=True,
    name=None,
):
    if conv_shortcut:
        shortcut = tf.keras.layers.Conv2D(
            filters,
            1,
            strides=stride,
            use_bias=False,
            kernel_initializer=tf.keras.initializers.LecunUniform(),
            name=name + "_0_conv",
        )(x)
        shortcut = tf.keras.layers.BatchNormalization(
            epsilon=1.001e-5, name=name + "_0_bn"
        )(shortcut)
    else:
        shortcut = x

    x = tf.keras.layers.Conv2D(
        filters,
        kernel_size,
        strides=stride,
        padding="SAME",
        use_bias=False,
        kernel_initializer=tf.keras.initializers.LecunUniform(),
        name=name + "_1_conv",
    )(x)
    x = tf.keras.layers.BatchNormalization(
        epsilon=1.001e-5, name=name + "_1_bn"
    )(x)
    x = tf.keras.layers.Activation("relu", name=name + "_1_relu")(x)

    x = tf.keras.layers.Conv2D(
        filters,
        kernel_size,
        padding="SAME",
        use_bias=False,
        kernel_initializer=tf.keras.initializers.LecunUniform(),
        name=name + "_2_conv",
    )(x)
    x = tf.keras.layers.BatchNormalization(
        epsilon=1.001e-5, name=name + "_2_bn"
    )(x)

    x = tf.keras.layers.Add(name=name + "_add")([shortcut, x])
    x = tf.keras.layers.Activation("relu", name=name + "_out")(x)
    return x


def stack0(x, filters, blocks, stride1=2, name=None):
    x = block0(x, filters, stride=stride1, name=name + "_block1")
    for i in range(2, blocks + 1):
        x = block0(
            x,
            filters,
            conv_shortcut=False,
            name=name + "_block" + str(i),
        )
    return x


def stack_fn(x):
    x = stack0(x, 64, 2, stride1=1, name="conv2")
    x = stack0(x, 128, 2, name="conv3")
    x = stack0(x, 256, 2, name="conv4")
    return stack0(x, 512, 2, name="conv5")



from matplotlib import pyplot as plt
from tensorflow import Tensor
import tensorflow as tf
from typing import Tuple


def visualize_views(views: Tensor,
                    labels: Tensor = None,
                    predictions: Tensor = None,
                    num_imgs: int = None,
                    views_per_col: int = 4,
                    fig_size: Tuple[int, int] = (24, 4),
                    max_pixel_value: float = 1.0,
                    min_pixel_value: float = 0.0):
    """Display side by side different image views with labels, and predictions

    Args:
        views: Aray of views
        predictions: model output.
        labels: image labels
        num_imgs: number of images to use.
        views_per_col: Int, number of images in one row. Defaults to 3.
        max_pixel_value: Max expected value for a pixel. Used to scale the image
          between [0,1].
        min_pixel_value: Min expected value for a pixel. Used to scale the image
          between [0,1].

    Returns:
        None.
    """
    num_views = len(views)
    num_imgs = num_imgs if num_imgs else len(views[0])
    num_col = views_per_col
    num_row = num_imgs // num_col
    num_row = num_row + 1 if num_imgs % num_col else num_row

    # Plot the images
    fig, axes = plt.subplots(num_row, num_col, figsize=fig_size)
    for i in range(num_imgs):

        # If the number of rows is 1, the axes array is one-dimensional
        if num_row == 1:
            ax = axes[i % num_col]
        else:
            ax = axes[i // num_col, i % num_col]

        scale = abs(max_pixel_value - min_pixel_value)
        pair = [(views[j][i] - min_pixel_value) / scale for j in range(num_views)]
        ax.imshow(tf.concat(pair, axis=1))
        ax.set_axis_off()

        label = labels[i] if labels else i

        if predictions:
            ax.set_title("Label: {} | Pred: {:.5f}".format(label,
                                                           predictions[i][0]))
        elif labels:
            ax.set_title("Label: {}".format(label))

        ax.set_xticks([])
        ax.set_yticks([])
    plt.show()
